/*
 * config.h
 *
 * In this file, the properties of the simulated system and the technique how
 * the simulation will be done is specified
 *
 *  Created on: Nov 4, 2009
 *      Author: blockben
 *
 * Copyright (C) 2010 Benjamin Block (http://www.lhyanor.org), 
 *                    Tobias Preis (http://tobiaspreis.de)
 *
 * This program is free software; you can redistribute it and/or 
 * modify it under the terms of the GNU General Public License 
 * as published by the Free Software Foundation; either version 
 * 3 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the 
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public 
 * License along with this program; if not, see 
 * http://www.gnu.org/licenses/.
 */

#ifndef CONFIG_H_
#define CONFIG_H_

/* Multi GPU information
 *
 * This Simulation package provides a multi GPU communication implementation using
 * MPI. When the MULTI_GPU flag is set, DEVICE_X * DEVICE_Y GPUs process a grid of
 * DEVICE_X * DEVICE_Y lattices
 * */
//#define MULTI_GPU
#define DEVICES_X 1
#define DEVICES_Y 1

/* The block size effectively defines the size of the spin field.
 * Current CUDA enabled graphics hardware poses a maximum block size of 512.
 * If bigger spin fields are needed, the BLOCK_DIVISION flag has to be set
 */
#define BLOCK_DIM 8

#define FIELD_DIM 128

/* If this flag is set, after each temperature step the final configuration will be written
 * into a TGA file, for debugging purposes
 */
#define DUMP_MEMORY

/* If flag is set, a CPU reference solution will be calculated and timed
 *
 */
#define CPU_REFERENCE


//#define GPU_REDUCTION

/* Define the temperature range to be sampled */
#define T_START 3.0
#define T_FACTOR 0.9
#define T_END 2.0

/* How many samples to collect for each temperature */
#define GLOBAL_ITERATIONS 1000


#endif /* CONFIG_H_ */
