/*
 * core.cpp
 *
 *  Created on: Oct 8, 2009
 *      Author: Benjamin Block

 * Copyright (C) 2010 Benjamin Block (http://www.lhyanor.org), 
 *                    Tobias Preis (http://tobiaspreis.de)
 *
 * This program is free software; you can redistribute it and/or 
 * modify it under the terms of the GNU General Public License 
 * as published by the Free Software Foundation; either version 
 * 3 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the 
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public 
 * License along with this program; if not, see 
 * http://www.gnu.org/licenses/.
 */

#include <cstdio>
#include <cmath>
#include <cstdlib>
#include <cstring>
#include <cstdarg>

#include "core.h"
#include "gpu_interface.cuh"
#include "mpi_communication.h"

int dump_ctr = 0;

struct rgb_dot
{
  unsigned char b, g, r;
};

#define _S(meta,x,y) (((meta & ( 1 << ((y) * 4 + (x)))) != 0)? 1 : -1)

int get_spin(unsigned short *field, int x, int y)
{
  int x_0 = x / 4;
  int y_0 = y / 4;

  unsigned short meta_spin = field[y_0 * (n/4) + x_0];
  return ((meta_spin & ( 1 << ((y % 4) * 4 + (x % 4)))) != 0)? 1 : -1;
}

int get_border_spin(unsigned short *borders, int num, int x, int y)
{
  int x_0 = x / 4;

  unsigned short *border = (borders + num*(n/4));

  return ((border[x/4] & ( 1 << ( (y %4) * 4 + (x % 4)))) != 0)? 1 : -1;
}


void set_spin(unsigned short *field, int x, int y, int value)
{
  int x_0 = x / 4;
  int y_0 = y / 4;

  unsigned short *meta_spin = &field[y_0 * (n/4) + x_0];

  if (value == 1)
    *meta_spin |= ( 1 << ((y % 4) * 4 + (x % 4)));
  else
    *meta_spin &= ~( 1 << ((y % 4) * 4 + (x % 4)));
}

void dump_host_memory(IsingHostData *hdata)
{

 int i;
 int myrank = rank;

 char fileName[32];
 sprintf(fileName, "memorycontent_rank%i.tga", myrank);

 unsigned char header[18];

 typedef rgb_dot row[n+8];
 row *image = new row[n+8];

 header[0] = 0; // offset
 header[1] = 0; // colortype
 header[2] = 2; // imagetype
 header[3] = 0; // palette start
 header[5] = 0; // palette len
 header[7] = 0; // palette bits
 header[8] = 0; // image x origin
 header[9] = 0; // second byte
 header[10] = 0; // image y origin
 header[11] = 0; // second byte
 *(short*)(&header[12]) = n + 8; // x width
 *(short*)(&header[14]) = n + 8; // y width
 header[16] = (unsigned char) 24; // bpp
 header[17] = 0; // orientation

 memset(image, 0x00, sizeof(rgb_dot) * (n+8) * (n+8));

 for (i=0; i < n; i++)
 {
   if (_S((hdata->theirBorders[N_TOP*(n/4) + i/4]), i%4, 3) > 0) image[n+8-1][i+4].r = 255;
   else
     image[n+8-1][i+4].g = 255;
 }

 for (i=0; i < n; i++)
 {
   if (_S((hdata->myBorders[N_TOP*(n/4) + i/4]), i%4, 0) > 0) image[n+8-3][i+4].r = 255;
   else
     image[n+8-3][i+4].g = 255;
 }



 for(i=0;i<n;i++) {

   if (_S((hdata->theirBorders[N_LEFT*(n/4) + i/4]), 3, i%4) > 0) image[n+8 - 1 - ( i + 4) ][0].r = 255;
   else image[n+8 - 1 - (i + 4)][0].g = 255;

   if (_S((hdata->myBorders[N_LEFT*(n/4) + i/4]), 0, i%4) > 0) image[n+8 - 1 - (i + 4)][2].r = 255;
   else image[n+8 - 1 - (i + 4)][2].g = 255;

   for(int j=0;j<n;j++) {
     if(get_spin(hdata->metaSpins, j, i)>0) image[n+8 - 1 - (i + 4)][j+4].r = 255;
     else image[n+8 - 1 - (i + 4)][j+4].g = 255;

       }

   if (_S((hdata->myBorders[N_RIGHT*(n/4) + i/4]), 3, i%4) > 0) image[n+8 - 1 - ( i + 4) ][n + 8 - 3].r = 255;
   else image[n+8 - 1 - ( i + 4) ][n + 8 - 3].g = 255;

   if (_S((hdata->theirBorders[N_RIGHT*(n/4) + i/4]), 0, i%4) > 0) image[n+8 - 1 - ( i + 4) ][n + 8 - 1].r = 255;
   else image[n+8 - 1 - ( i + 4) ][n + 8 - 1].g = 255;

   }

for (i=0; i < n; i++)
{
  if (_S((hdata->myBorders[N_BOTTOM*(n/4) + i/4]), i%4, 3) > 0) image[2][i+4].r = 255;
  else
    image[2][i+4].g = 255;
}

for (i=0; i < n; i++)
{
  if (_S((hdata->theirBorders[N_BOTTOM*(n/4) + i/4]), i%4, 0) > 0) image[0][i+4].r = 255;
  else
    image[0][i+4].g = 255;
}

 FILE * file = fopen(fileName, "wb");

 fwrite(header, sizeof(char) * 18, 1, file);
 fwrite(image, sizeof(rgb_dot) * (n+8) * (n+8), 1, file);
 fclose(file);

 delete[] image;
}


void log_write(const char *fmt, ...)
{
  if (rank == 0)
    {

    va_list args;

    printf("@ ");

    va_start(args, fmt);
    vprintf(fmt, args);
    va_end(args);

    printf("\n");
    }
}

void log_write_all(const char *fmt, ...)
{
    va_list args;

    printf("@ [%i] ", rank);

    va_start(args, fmt);
    vprintf(fmt, args);
    va_end(args);

    printf("\n");
}

/****
 *
 *  CPU function
 *
 */

void cpu_alloc_memory(short **S)
{
  *S = new short[N* DEVICES_X * DEVICES_Y];
}

void cpu_free_memory(short *S)
{
  delete[] S;
}
void cpu_reference(short* S, int entries)
{
  double E[entries];

#define N_whole (N * DEVICES_X * DEVICES_Y)
#define n_whole (n * DEVICES_X)

  int random=23;
  int num_entries=0;

  for(double t=T_START;t>=T_END;t=t*T_FACTOR) {
    double avg_H=0;
    double exp_dH_4=exp(-(4.0)/t);
    double exp_dH_8=exp(-(8.0)/t);

    for(int global_iteration=0;global_iteration<GLOBAL_ITERATIONS;++global_iteration) {

        //Magnetisation
        double H=0;
        for(int x=0;x<N_whole;++x) {
          H+=S[x];
        }
        avg_H+=H/N_whole;

      for(int x=0;x<n_whole;++x) {
        for(int y=0;y<n_whole;++y) {
          if((y*(n_whole+1)+x)%2==0) {
            int xl=x-1,yl=y,xu=x,yu=y-1,xr=x+1,yr=y,xd=x,yd=y+1;
            if(x==0) {
              xl=n_whole-1;
            }
            else if(x==n_whole-1) {
              xr=0;
            }
            if(y==0) {
              yu=n_whole-1;
            }
            else if(y==n_whole-1) {
              yd=0;
            }

            //Initial local energy
            int dH=2*S[y*n_whole+x]*(
                               S[yl*n_whole+xl]+
                               S[yr*n_whole+xr]+
                               S[yu*n_whole+xu]+
                               S[yd*n_whole+xd]
                               );

            if(dH==4) {
              random=RANDOM_A*random+RANDOM_B;
              if(fabs(random*4.656612e-10)<exp_dH_4) {
                S[y*n_whole+x]=-S[y*n_whole+x];
              }
            }
            else if(dH==8) {
              random=RANDOM_A*random+RANDOM_B;
              if(fabs(random*4.656612e-10)<exp_dH_8) {
                S[y*n_whole+x]=-S[y*n_whole+x];
              }
            }
            else {
              S[y*n_whole+x]=-S[y*n_whole+x];
            }
          }
        }
      }

      for(int x=0;x<n_whole;++x) {
        for(int y=0;y<n_whole;++y) {
          if((y*(n_whole+1)+x)%2==1) {
            int xl=x-1,yl=y,xu=x,yu=y-1,xr=x+1,yr=y,xd=x,yd=y+1;
            if(x==0) {
              xl=n_whole-1;
            }
            else if(x==n_whole-1) {
              xr=0;
            }
            if(y==0) {
              yu=n_whole-1;
            }
            else if(y==n_whole-1) {
              yd=0;
            }

            //Initial local energy
            int dH=2*S[y*n_whole+x]*(
                               S[yl*n_whole+xl]+
                               S[yr*n_whole+xr]+
                               S[yu*n_whole+xu]+
                               S[yd*n_whole+xd]
                               );

            if(dH==4) {
              random=RANDOM_A*random+RANDOM_B;
              if(fabs(random*4.656612e-10)<exp_dH_4) {
                S[y*n_whole+x]=-S[y*n_whole+x];
              }
            }
            else if(dH==8) {
              random=RANDOM_A*random+RANDOM_B;
              if(fabs(random*4.656612e-10)<exp_dH_8) {
                S[y*n_whole+x]=-S[y*n_whole+x];
              }
            }
            else {
              S[y*n_whole+x]=-S[y*n_whole+x];
            }
          }
        }
      }
    }
    E[num_entries]=avg_H/GLOBAL_ITERATIONS;
    num_entries++;
  }
}

void title()
{
  printf("\n");
  printf("<<<<<< Multi GPU Implementation of the 2D Ising Model >>>>>>\n\n");

  printf(" Block Linear Dimension: %d \n", BLOCK_DIM);

  printf(" Field Linear Dimension: %d Blocks \n", FIELD_DIM);

#ifdef MULTI_GPU
  printf(" Multi GPU: %d x %d GPUs \n", DEVICES_X, DEVICES_Y);
#endif

#ifdef GPU_REDUCTION
  printf(" Reduction is fully done on GPU\n");
#endif

#ifdef STORE_FIELD
  printf(" Storing magnetisation field for faster reduction\n");
#endif

  printf(" Number of Spins: %d x %d = %d \n",(n) * DEVICES_X, (n) * DEVICES_Y, N* DEVICES_X * DEVICES_Y);
  printf(" Start Temperature: %f \n",T_START);
  printf(" Decreasing Factor: %f \n",T_FACTOR);
  printf(" Final Temperature: %f \n",T_END);
  printf(" Global Iterations: %d \n",GLOBAL_ITERATIONS);

  printf("\n");
}

int main(int argc, char ** argv)
{
#ifdef MULTI_GPU
  mpi_init(argc, argv);
#endif

  /* structures storing host and device data */
  IsingHostData hdata;
  IsingDeviceData ddata;

  if (rank == 0)
    title();

  /* set up gpu */
  int ndevices = gpu_query(rank);

  /* check if user is sane */
  if ((nthreads != DEVICES_X*DEVICES_Y ))
    {
    log_write("CPU Threads and requested or installed devices do not match!");
    log_write("Threads: %i, Requested Devices: %i, Installed Devices: %i", nthreads, DEVICES_X*DEVICES_Y, ndevices);
    return 1;
    }

  if (nthreads > ndevices)
  {
    log_write("Warning: Simulating more subsystems than GPUs available, some GPUs will be used for more than one system - this is not fatal but inefficient.");
  }

  /* define where we are in the grid, and what our neighbours are */
  int dx = rank%DEVICES_X;
  int dy = rank/DEVICES_X;

  int neighbours[4];

  /* define neighbours */
  neighbours[N_LEFT] = dy * DEVICES_X + ((dx-1)+DEVICES_X) % DEVICES_X;
  neighbours[N_RIGHT] = dy * DEVICES_X + ((dx+1)+DEVICES_X) % DEVICES_X;
  neighbours[N_TOP] = ((dy-1)+DEVICES_Y) % DEVICES_Y * DEVICES_X + dx;
  neighbours[N_BOTTOM] = ((dy+1) + DEVICES_Y) % DEVICES_Y * DEVICES_X + dx;

  log_write("--------------------------------- GPU ---------------------------------");

  srand48(543);

  //Allocate and init host memory for output arrays
  int num_entries=0;
  for(float t=T_START; t>=T_END; t=t*T_FACTOR) num_entries++;

  float temperatures[num_entries];
  double m[num_entries];
  double m4[num_entries];
  double m2[num_entries];

  num_entries=0;
  for(float t=T_START; t>=T_END; t=t*T_FACTOR) {
    temperatures[num_entries]=t;
    num_entries++;
  }

  int i;

  /* allocate device and host memory */
  gpu_alloc_memory(&ddata, &hdata);

  for(i=0;i<N/16;i++) {
    hdata.metaSpins[i] = 0;
  }

  /* create random data for gpu usage */
  hdata.randomData[0] = 1+rank*12345;

  for(i=1;i<NUM_THREADS;i++)
    hdata.randomData[i]=(16807) *hdata.randomData[i-1];

  // prepare my borders
  memcpy(&hdata.myBorders[N_TOP*(n/4)], hdata.metaSpins, sizeof(short) * (n/4));
  memcpy(&hdata.myBorders[N_BOTTOM*(n/4)], &hdata.metaSpins[(N/16) - (n/4)], sizeof(short) * (n/4));
  for(i = 0; i < (n/4); i++)
    {
      hdata.myBorders[N_LEFT*(n/4)+ i] = hdata.metaSpins[i*(n/4)];
      hdata.myBorders[N_RIGHT*(n/4)+ i] = hdata.metaSpins[(i+1)*(n/4)-1];
    }


#ifdef MULTI_GPU
  mpi_exchange_borders(&hdata, rank, neighbours);
#endif

  //Create and start timer
  float gpu_sum=0;

  gpu_timer_start();



  float gpu_dt_malloc = gpu_timer_stop();
  gpu_sum+=gpu_dt_malloc;
  log_write(" Processing time on GPU for allocating: %f (ms)",gpu_dt_malloc);


  gpu_timer_start();
  gpu_data_deposit(&ddata, &hdata);

  float gpu_dt_mem=gpu_timer_stop();
  gpu_sum+=gpu_dt_mem;
  log_write(" Processing time on GPU for memory transfer: %f (ms)",gpu_dt_mem);

  //Calc energy
  num_entries=0;

  gpu_timer_start();

  for(float t=T_START;t>=T_END;t=t*T_FACTOR) {

    double avg_m, avg_m2, avg_m4;
    avg_m = avg_m2 = avg_m4 = 0.0f;
    
    ddata.exp_dH_4=exp(-(4.0)/t);
    ddata.exp_dH_8=exp(-(8.0)/t);


    for(int global_iteration=0;global_iteration<GLOBAL_ITERATIONS;global_iteration++) {
      
#ifdef MULTI_GPU
      /* multi gpu mainloop */
      gpu_borders_deposit(&ddata, &hdata);
      gpu_kernel_call(&ddata, t, true);
      gpu_borders_retrieve(&ddata, &hdata);

      mpi_exchange_borders(&hdata, rank, neighbours);

      gpu_borders_deposit(&ddata, &hdata);
      gpu_kernel_call(&ddata, t, false);
      gpu_borders_retrieve(&ddata, &hdata);

      mpi_exchange_borders(&hdata, rank, neighbours);
#else
      /* single GPU implementation */
      gpu_kernel_call(&ddata, t, true);

      gpu_kernel_call(&ddata, t, false);
#endif

      /* diagnose */
#ifdef DUMP_MEMORY
      if (dump_ctr == GLOBAL_ITERATIONS-1)
      {
              gpu_spins_retrieve(&ddata,&hdata);
              //if (n < 512)
                dump_host_memory(&hdata);
              dump_ctr = 0;
      }
      dump_ctr++;
#endif

#ifdef EQUILIBRATE_FIRST
      if (global_iteration > GLOBAL_ITERATIONS * 0.5)
        {
#endif

      /* magnetisation */
      gpu_out_retrieve(&ddata, &hdata);

#ifdef GPU_REDUCTION
      /* magnetisation has already been reduced to the first element
       * of the out array
       */
      int magnetisation=hdata.out[0];
#else
      /* do last step of reduction on the cpu */
      int magnetisation=0;

      for(int i=0; i< FIELD_DIM * FIELD_DIM;i++) {
        magnetisation+=hdata.out[i];
      }
#endif

      double this_m = (double)magnetisation/N;

#ifdef MULTI_GPU
      mpi_reduce_result(&this_m, 1);
#endif

      double this_m2 = this_m*this_m;
      double this_m4 = this_m2*this_m2;

      avg_m += (double) this_m;
      avg_m2 += (double) this_m2;
      avg_m4 += (double) this_m4;

#ifdef EQUILIBRATE_FIRST
      }
#endif
    }

#ifdef EQUILIBRATE_FIRST
    m[num_entries] =  avg_m * 2 / GLOBAL_ITERATIONS;
    m2[num_entries] = avg_m2 * 2 / GLOBAL_ITERATIONS;
    m4[num_entries] = avg_m4 * 2 / GLOBAL_ITERATIONS;
#else
    m[num_entries] =  avg_m / GLOBAL_ITERATIONS;
    m2[num_entries] = avg_m2 / GLOBAL_ITERATIONS;
    m4[num_entries] = avg_m4 / GLOBAL_ITERATIONS;
#endif

#ifdef DUMP_MEMORY
    if (rank == 0)
      printf("%f %lf %lf %lf %lf\n", temperatures[num_entries],m[num_entries], m2[num_entries], m4[num_entries], (1.0 - (double)m4[num_entries]/(3.0*(double)m2[num_entries]*m2[num_entries])));
#endif
    num_entries++;
  }



  float gpu_dt_main=gpu_timer_stop();
  gpu_sum+=gpu_dt_main;
  log_write(" Processing time on GPU for main function: %f (ms)",gpu_dt_main);
  log_write(" Total processing time on GPU: %f (ms)",gpu_sum);

  log_write("");

  if (rank == 0)
  {
    for (i = 0; i < num_entries; i++)
      {
        printf("%f %lf %lf %lf %lf\n", temperatures[i],m[i], m2[i], m4[i], (1.0 - (double)m4[i]/(3.0*(double)m2[i]*m2[i])));
      }
  }



#ifdef CPU_REFERENCE
  if (rank == 0)
    {
    log_write("--------------------------------- CPU ---------------------------------");

    gpu_free_memory(&ddata, &hdata);

    short *spins;
    cpu_alloc_memory(&spins);

    gpu_timer_start();

    cpu_reference(spins,num_entries);

    float cpu_sum=gpu_timer_stop();



    printf(" Total processing time on CPU: %f (ms) \n",cpu_sum);
    printf("\n Speedup: %fX \n\n",(cpu_sum/gpu_sum));

    cpu_free_memory(spins);
    }
#endif

  return 0;
}
