/*
 * core.h
 *
 * This file defines the basic data structures and definitions used throughout the program
 *
 *  Created on: Oct 8, 2009
 *      Author: Benjamin Block
 *
 * Multi-GPU Implementation of the 2D Ising Model (Shared Memory Version)
 *
 * Copyright (C) 2010 Benjamin Block (http://www.lhyanor.org), 
 *                    Tobias Preis (http://tobiaspreis.de)
 *
 * This program is free software; you can redistribute it and/or 
 * modify it under the terms of the GNU General Public License 
 * as published by the Free Software Foundation; either version 
 * 3 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the 
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public 
 * License along with this program; if not, see 
 * http://www.gnu.org/licenses/.
 */

#ifndef CORE_H_
#define CORE_H_

#include "config.h"

/* The geometry of the lattices (on each GPU) - dont modify this! 
 * N is the total spin number, n is the linear lattice size 
 */
#define N (BLOCK_DIM * BLOCK_DIM * 16 * FIELD_DIM * FIELD_DIM)
#define n (BLOCK_DIM * 4 * FIELD_DIM)

/* The data structure storing all information about the spin field needed on the host side
 * A similar structure for the GPU data is defined in gpu_interface.cuh
 */
struct IsingHostData
{
  /* (4x4 spins) in each short */
  unsigned short *metaSpins;
  // the random seeds passed to the graphics hardware
  int *randomData;
  // this is used for reduction
  int *out;
  // i need to have information about the borders of the others
  unsigned short *theirBorders;
  // my borders have to get prepared to be send to others
  unsigned short *myBorders;

};

/* to tell the neighbours appart */
#define N_LEFT 0
#define N_RIGHT 1
#define N_BOTTOM 2
#define N_TOP 3

/* constants for the random number generator */
#define RANDOM_A 1664525
#define RANDOM_B 1013904223

/* function definitions */
void log_write_all(const char *fmt, ...);
void log_write(const char *fmt, ...);
void dump_host_memory(IsingHostData *hdata);

#endif /* MAIN_H_ */
